import threading

from feathery.constants import (
    API_URL,
    REFRESH_INTERVAL,
    REQUEST_TIMEOUT,
    POLL_FREQ_SECONDS,
)
from feathery.polling import PollingThread
from feathery.utils import fetch_and_return_settings


class FeatheryClient:
    def __init__(self, sdk_key):
        """Sets the SDK key and spins up an asynchronous setting polling job.
        :param string sdk_key: the new SDK key
        """

        self.sdk_key = sdk_key
        self.settings = {}

        self.api_url = API_URL
        self.refresh_interval = REFRESH_INTERVAL
        self.request_timeout = REQUEST_TIMEOUT
        self._lock = threading.Lock()

        self.settings = fetch_and_return_settings(self.sdk_key)

        # Start periodic job
        self.scheduler = PollingThread(
            features=self.settings,
            sdk_key=self.sdk_key,
            interval=POLL_FREQ_SECONDS,
            lock=self._lock,
        )
        self.scheduler.start()

        self.is_initialized = True

    def variation(self, setting_key, default_value, user_key):
        """
        Checks the setting value for a user.  If the user and setting exist,
        return variant.
        Notes:
        * If client hasn't been initialized yet or an error occurs, flat will
        default to false.
        :param setting_key: Name of the setting
        :param default_value: Default value for the setting.
        :param user_key: Unique key belonging to the user.
        :return: Dict with variant and setting status.
        """

        variant = default_value

        if self.is_initialized:
            self._lock.acquire()
            if setting_key in self.settings:
                if user_key in self.settings[setting_key]["overrides"]:
                    variant = self.settings[setting_key]["overrides"][user_key]
                else:
                    variant = self.settings[setting_key]["value"]
            self._lock.release()

        return variant

    def destroy(self):
        """
        Gracefully shuts down the Feathery client by stopping jobs, stopping
        the scheduler, and deleting the cache.
        :return:
        """
        self.scheduler.stop()



import requests

from feathery.constants import API_URL, REQUEST_TIMEOUT


def fetch_and_return_settings(sdk_key: str) -> dict:
    new_settings = get_settings_json(sdk_key)

    new_dict = {}
    for item in new_settings:
        name = item["key"]
        overrides = {
            override_item["user_key"]: override_item["user_value"]
            for override_item in item["overrides"]
        }
        new_dict[name] = {
            "value": item["value"],
            "datatype": item["datatype"],
            "overrides": overrides,
        }
    return new_dict


def get_settings_json(sdk_key: str) -> dict:
    """
    Retrieves settings from Feathery central server.
    Notes:
    * If unsuccessful (i.e. not HTTP status code 200), exception will be caught and logged.
        This is to allow "safe" error handling if unleash server goes down.
    :return: Configurations if successful, empty dict if not.
    """
    headers = {"Authorization": "Token " + sdk_key}
    resp = requests.get(API_URL, headers={**headers}, timeout=REQUEST_TIMEOUT)
    if resp.status_code != 200:
        return {}
    return resp.json()



import copy
import threading
import time

from feathery.utils import fetch_and_return_settings


class PollingThread(threading.Thread):
    def __init__(self, features, sdk_key, interval, lock):
        threading.Thread.__init__(self)
        self._running = False
        self.features = features
        self.sdk_key = sdk_key
        self.interval = interval
        self.lock = lock

    def run(self):
        if not self._running:
            self._running = True
            while self._running:
                start_time = time.time()
                try:
                    all_data = fetch_and_return_settings(self.sdk_key)
                    self.lock.aquire()
                    self.features = copy.deepcopy(all_data)
                    self.lock.release()
                except Exception:
                    pass

                elapsed = time.time() - start_time
                if elapsed < self.interval:
                    time.sleep(self.interval - elapsed)

    def stop(self):
        self._running = False



"""
SDK public entry points are contained here. This design enforces the singleton
model -- only one instance of `feathery` can exist at any time.
"""
from feathery.client import FeatheryClient

__client = None


def set_sdk_key(sdk_key):
    """Sets the SDK auth key and initializes the shared client instance.
    If the shared client instance is already initialized, this function no-ops.
    :param string sdk_key: the new SDK key
    """
    global __client
    if __client is None:
        __client = FeatheryClient(sdk_key)


def get():
    """Returns the shared SDK client instance. To use the SDK as a singleton,
    first make sure you have called :func:`feathery.set_sdk_key()` at
    startup time. Then ``get()`` will then return the same shared
    :class:`feathery.client.FeatheryClient` instance each time. If
    :func:`feathery.set_sdk_key()` has not been called, `None` will be
    returned.
    :rtype: feathery.client.FeatheryClient
    """
    global __client
    return __client



import setuptools

setuptools.setup(
    name="feathery", # Replace with your own username
    version="0.0.1",
    author="Markie Wagner, Peter Dun",
    author_email="me@markiewagner.com",
    description="A Python SDK for Feathery.",
    url="https://github.com/bo-dun-1/feathery-python-server-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)